/*
Copyright 2018 Heptio Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package app

import (
	"os"
	"time"

	"github.com/pkg/errors"
	"github.com/spf13/cobra"
	"github.com/vmware-tanzu/sonobuoy/pkg/client"
	"github.com/vmware-tanzu/sonobuoy/pkg/errlog"
	"github.com/vmware-tanzu/sonobuoy/pkg/features"
)

type deleteFlags struct {
	namespace  string
	rbacMode   RBACMode
	deleteAll  bool
	wait       int
	kubeconfig Kubeconfig
	waitOutput WaitOutputMode
	dryRun     bool
}

func NewCmdDelete() *cobra.Command {
	f := deleteFlags{}
	cmd := &cobra.Command{
		Use:   "delete",
		Short: "Deletes Kubernetes resources that were generated by a Sonobuoy run",
		Run:   deleteSonobuoyRun(&f),
		Args:  cobra.ExactArgs(0),
	}

	AddKubeconfigFlag(&f.kubeconfig, cmd.Flags())
	AddNamespaceFlag(&f.namespace, cmd.Flags())
	AddRBACModeFlags(&f.rbacMode, cmd.Flags(), DetectRBACMode)
	AddDeleteAllFlag(&f.deleteAll, cmd.Flags())
	AddDryRunFlag(&f.dryRun, cmd.Flags())
	AddDeleteWaitFlag(&f.wait, cmd.Flags())
	if features.Enabled(features.WaitOutputProgressByDefault) {
		AddWaitOutputFlag(&f.waitOutput, cmd.Flags(), ProgressOutputMode)
	} else {
		AddWaitOutputFlag(&f.waitOutput, cmd.Flags(), SilentOutputMode)
	}
	return cmd
}

func deleteSonobuoyRun(f *deleteFlags) func(cmd *cobra.Command, args []string) {
	return func(cmd *cobra.Command, args []string) {
		sbc, err := getSonobuoyClientFromKubecfg(f.kubeconfig)
		if err != nil {
			errlog.LogError(errors.Wrap(err, "could not create sonobuoy client"))
			os.Exit(1)
		}

		kc, err := sbc.Client()
		if err != nil {
			errlog.LogError(err)
			os.Exit(1)
		}

		rbacEnabled, err := f.rbacMode.Enabled(kc)
		if err != nil {
			errlog.LogError(errors.Wrap(err, "couldn't detect RBAC status"))
			os.Exit(1)
		}

		deleteCfg := &client.DeleteConfig{
			Namespace:  f.namespace,
			EnableRBAC: rbacEnabled,
			DeleteAll:  f.deleteAll,
			Wait:       time.Duration(f.wait) * time.Minute,
			WaitOutput: f.waitOutput.String(),
			DryRun:     f.dryRun,
		}

		if err := sbc.Delete(deleteCfg); err != nil {
			errlog.LogError(errors.Wrap(err, "failed to delete sonobuoy resources"))
			os.Exit(1)
		}
	}
}
